# CritLink Mediator: Document and link database package.
# Implements the M1db class.  This object is the interface to a
# database on disk, and provides the ability to write forward
# links into the database and to read forward links, backward
# links, or metainformation for a particular document.
#
# Exports: ->fwdlinks($url), ->backlinks($url),
#          ->docinfo($url), ->store($source, @links)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1db;
use M1ui;
use M1url;
use M1link;
use M1term;
use strict;

my $DBPERMISSIONS = 0755;              # database is world-readable 
my $NOTIFY_TEMPLATE = "$M1config::docpath/cl-notify.msg";
my $NOTIFYREMOVE_TEMPLATE = "$M1config::docpath/cl-notifyremove.msg";
my $MONITOR_TEMPLATE = "$M1config::docpath/cl-monitor.msg";

sub new
    {
    my ($class, $root) = @_;           # string: root path of link database
    $root =~ s/\/+$//;
    my $self = {'root' => $root,
                'sroot' => $root . '/source',
                'troot' => $root . '/target',
                'cache' => {} };
    mkdir($self->{'root'}, $DBPERMISSIONS);
    mkdir($self->{'sroot'}, $DBPERMISSIONS);
    mkdir($self->{'troot'}, $DBPERMISSIONS);
    return bless($self, $class);
    }

sub docinfo
    {
    my ($self, $url) = @_;             # M1url: URL of a document
    return {} if ref($url) ne 'M1url';
    my $file = $url->filename($self->{'sroot'});
    return ${$self->{'cache'}}{$file} ||= &getdocinfo($file);

    sub getdocinfo
        {
        my ($file, $dict) = ($_[0], {});
        open(FILE, $file);
        while (<FILE> =~ /:\s*(.*)/)
            { $dict->{$`} = $1; }
        close(FILE);
        return $dict;
        }
    }

sub fwdlinks
    {
    my ($self, $url) = @_;             # M1url: URL of source document
    my $file = $url->filename($self->{'sroot'});
    my @links;

    open(SOURCE, $file);
    1 while <SOURCE> =~ /:/;
    while (<SOURCE>)
        {
        chop;
        my ($frag, $target, $title, $rel, $rev) = split("\t");
        next if $target eq '';
        $target = new M1url($target);
        my $source = $frag ? $url->relative("#$frag") : $url;
        push(@links, new M1link($source, $target, $title, $rel, $rev));
        }
    close(SOURCE);
    return @links;
    }

sub backlinks
    {
    my ($self, $url) = @_;             # M1url: URL of target document
    my $file = $url->filename($self->{'troot'});
    my (@links, @updates);
    $url = new M1url($url->canon('nofrag'));

    open(TARGET, $file) or return ();
    while (<TARGET>)
        {
        chop;
        my ($update, $source, $frag, $title, $rel, $rev) = split("\t");
        next if $source eq '';
        $source = new M1url($source);
        my $info = $self->docinfo($source);
        if ($update == $info->{'update'})
            {
            my $target = $frag ? $url->relative("#$frag") : $url;
            push(@links, new M1link($source, $target, $title, $rel, $rev));
            push(@updates, $update);
            }
        }

    &prepare($file);
    if (mkdir("$file\"lock", 0700))    # simple atomic operation for locking
        {
        &M1term::msg("rewriting links with target " . $url->canon());
        open(TARGET, ">$file") || &M1term::msg("could not write $file: $@");
        my $link;
        foreach $link (@links)
            {
            print TARGET join("\t", shift(@updates),
                                    $link->{'source'}->canon(),
                                    $link->{'target'}->{'frag'},
                                    $link->{'title'},
                                    $link->{'rel'},
                                    $link->{'rev'}) . "\n";
            }
        rmdir("$file\"lock");
        }
    else
        { &M1term::msg("could not lock target for " . $url->canon()); }
    close(TARGET);
    return @links;
    }

sub store
    {
    my ($self, $url, @links) = @_;     # M1url: URL of source document
    my $file = $url->filename($self->{'sroot'});
    my $info = $self->docinfo($url);
    my $lockdir = $file . '"lock';
    my ($locktry, $key, $link);

    &prepare($lockdir);
    if ((stat($lockdir))[9] < time() - 20 * 60)  # remove stale lock
        { rmdir($lockdir); }

    while (!mkdir($lockdir, 0700))     # simple atomic operation for locking
        {
        if ($locktry++ > 10) { return 0; }
        select('', '', '', 0.3);
        }

    $info->{'update'}++;
    &M1term::msg("writing links with source " . $url->canon() .
                 " (" . (@links+0) . " links)");

    my %oldlinks;
    foreach $link ($self->fwdlinks($url))
        { $oldlinks{$link->{'target'}->canon('nofrag')} = $link; }

    &prepare($file);
    open(SOURCE, ">$file") || &M1term::msg("cannot write $file: $!");
    foreach $key (keys %$info)
        {
        next if $$info{$key} eq '';
        $$info{$key} =~ s/\s*\n\s*/ /g;
        print SOURCE "$key: $$info{$key}\n";
        }
    print SOURCE "\n";

    my %seen;
    foreach $link (@links)
        { 
        print SOURCE join("\t", $link->{'source'}->{'frag'},
                                $link->{'target'}->canon(),
                                $link->{'title'},
                                $link->{'rel'},
                                $link->{'rev'}) . "\n";

        my $file = $link->{'target'}->filename($self->{'troot'});
        my $entry = join("\t", $info->{'update'},
                               $link->{'source'}->canon(),
                               $link->{'target'}->{'frag'},
                               $link->{'title'},
                               $link->{'rel'},
                               $link->{'rev'});

        &M1term::msg("writing link with target " . $link->{'target'}->canon());
        &M1term::dbg("    " . $link->repr());
        &prepare($file);
        open(TARGET, ">> $file") || &M1term::msg("cannot append to $file: $!");
        print TARGET "$entry\n";
        close(TARGET);

        if ($info->{'trigger'} =~ /never/)
            { next; }

        # will not notify on new links from other places in the
        # same page to other places in the same target URL
        my $key = $link->{'target'}->canon('nofrag');
        if (! $seen{$key})
            {
            $seen{$key} = 1;
            if (! $oldlinks{$key})
                { $self->notify($link); }
            }
        }

    if ($info->{'notrigger'} !~ /never/)
        {
        foreach $key (keys %oldlinks)
            {
            if (! $seen{$key})
                { $self->notify($oldlinks{$key}, 'remove'); }
            }
        }
        
    rmdir($lockdir);
    return 1;
    }

sub notify
    {
no strict 'vars';

    sub sendmessage
        {
        local($template, $address,
              $source, $sourcetitle, $target, $targettitle,
              $author, $type, $gran) = @_;

        &M1term::msg("sending notification message to $address");

        # extra logging while we are trying out this new feature
        $address =~ s/\s*$//;
        ($s, $m, $h, $d, $l, $y) = localtime(time());
        $stamp = sprintf("%02d-%02d-%02d %02d:%02d", $y%100, $l+1, $d, $h, $m);

        if (-x $M1config::mailbin)
            {
            open(MAIL, '|' . $M1config::sendmail);
            print MAIL &M1ui::render($template, 'M1config', 'M1db');
            close(MAIL);
            }
        }

    my ($self, $link, $remove) = @_;
    my $sourceinfo = $self->docinfo($link->{'source'});
    my $targetinfo = $self->docinfo($link->{'target'});

    my %linkdict;
    $linkdict{'type'} = $link->{'rev'};
    $linkdict{'gran'} = $link->{'target'}->{'frag'} ? 'fine' : 'coarse';
    $linkdict{'author'} =
        ($sourceinfo->{'authorlink'} =~ /mailto:(\S+)/ ?
         $1 : $sourceinfo->{'author'});
    $linkdict{'change'} = $remove ? 'removed' : 'new';

    my ($key, $expr, $term, $value);
    foreach $key (keys %linkdict)
        { $linkdict{$key} =~ tr/A-Z/a-z/; $linkdict{$key} =~ s/^\s*|\s*$//g; }

    if ($targetinfo->{'notify'} !~ /[^\s;]/i &&
        $targetinfo->{'monitor'} !~ /\S/) { return; }

    &M1term::msg($linkdict{'change'} . " link " . &repr::repr($link) .
                 " to document " . &repr::repr($targetinfo));

    if ($targetinfo->{'notify'} =~ /[^\s;]/)
        {
        my $exprvalue = 0;
        foreach $expr (split(';', $targetinfo->{'notify'}))
            {
            next unless $expr =~ /\S/;
            if ($expr =~ /^\s*all\s*$/)
                { $exprvalue = 1; last; }

            my $termvalue = 1;
            foreach $term (split(',', $expr))
                {
                next unless $term =~ /^\s*(\w+)\s*(not\s*)?\(([^>]*)\)\s*$/i;
                my ($key, $neg, $values) = ($1, $2, $3);
                $key =~ tr/A-Z/a-z/;
                $values =~ tr/A-Z/a-z/;

                my $matched = 0;
                foreach $value (split(' ', $values))
                    { if ($linkdict{$key} eq $value) { $matched = 1; } }

                if ($neg) { $matched = !$matched; }
                if (!$matched) { $termvalue = 0; }
                }

            if ($termvalue) { $exprvalue = 1; last; }
            }

        if ($exprvalue)
            {
            &M1term::msg("link satisfied notification filter!");
            if ($targetinfo->{'authorlink'} =~ /mailto:(\S+)/)
                {
                my $temp = $remove ? $NOTIFYREMOVE_TEMPLATE : $NOTIFY_TEMPLATE;
                &sendmessage($temp, $1,
                    $link->{'source'}->canon(), $sourceinfo->{'title'},
                    $link->{'target'}->canon(), $targetinfo->{'title'},
                    $linkdict{'author'}, $linkdict{'type'}, $linkdict{'gran'});
                }
            }
        }
    elsif ($targetinfo->{'monitor'} =~ /\S/ && !$remove)
        {
        my $addresses = '';

        foreach $address (split(' ', $targetinfo->{'monitor'}))
            {
            $address =~ s/[^a-zA-Z0-9_=\@%.-]//;
            next unless $address =~ /\@/ && $address =~ /\./;
            $addresses .= "$address ";
            }

        &sendmessage($MONITOR_TEMPLATE, $addresses,
            $link->{'source'}->canon(), $sourceinfo->{'title'},
            $link->{'target'}->canon(), $targetinfo->{'title'},
            $linkdict{'author'}, $linkdict{'type'}, $linkdict{'gran'});
        }
    }

# static helper method: create the chain of parent directories to a file
sub prepare
    {
    my @dirs = split('/', $_[0]);
    my ($dir, $path);
    pop(@dirs);

    foreach $dir (@dirs)
        {
        $path .= $dir;
        mkdir($path, 0755) if $path =~ m#[^/]#;
        $path .= '/';
        }
    }

1;
