# CritLink Mediator: Link finder package.
# Implements the M1finder class.  This sprocket examines tokens from
# the lexical analyzer and inserts indicators where links for the
# announced target URL are found according to a given link database.
#
# Exports: ->recv($token)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1finder;
use M1link;
use M1term;
use M1token;
use repr;
use strict;

@M1finder::ISA = 'sprocket';           # inherit from "sprocket"
$M1finder::linkid = 0;                 # unique identifier for each link found

sub new
    {
    my ($class, $next, $db, $sel) = @_;
    return bless({'next'    => $next,  # sprocket: the next sprocket
                  'db'      => $db,    # M1db: database to read from
                  'select'  => $sel,   # string: optional selected fragment id

                  'coarse'  => [],     # private: list of coarse links
                  'fine'    => {},     # private: fine links keyed by fragment
                  'partial' => {},     # private: map partial fragments to 1
                  'named'   => {},     # private: named fragments keyed by name
                  'words'   => [],     # private: word queue
                  'tags'    => [],     # private: tag queue
                  'pfrags'  => [],     # private: potential fragments
                  'sought'  => {} },   # private: not-yet-found fragments
                 $class);
    }

sub recv
    {
    my ($self, $token) = @_;
    my ($type, $data) = ($token->{'type'}, $token->{'data'});
    &M1term::dbg($token);

    if ($type eq 'status:url')         # get links for the target URL
        {
        $self->send($token);
        $self->readlinks($data);
        my $link;
        foreach $link (@{$self->{'coarse'}})
            { $self->send(new M1token('link:coarse', $link)); }
        return;
        }

    if ($type eq 'status:start')
        { $self->{'started'} = 1; }

    if (!$self->{'started'})
        { $self->send($token); return; }

    if ($type eq 'status:end')         # flush data and missing links
        {
        $self->flush();
        my ($frag, $link);
        my $fine = $self->{'fine'};
        my $named = $self->{'named'};
        foreach $frag (keys %{$self->{'sought'}})
            {
            foreach $link (@{$$fine{$frag}})
                { $self->send(new M1token('link:missing', $link)); }
            }
        foreach $frag (keys %{$self->{'named'}})
            { 
            foreach $link (@{$$named{$frag}})
                { $self->send(new M1token('link:missing', $link)); }
            }
        }

    if ($type =~ /^status:|^error:/ or !$self->{'fine'})
        { $self->send($token); return; }

    my $words = $self->{'words'};
    my $tags = $self->{'tags'};
    my $fine = $self->{'fine'};
    my $partial = $self->{'partial'};
    my $pfrags = $self->{'pfrags'};

    if ($type eq 'data:text')
        {
        my ($word, $punct, $keep, $discard);
        while ($data =~ /[a-zA-Z0-9']+/)          # find valid words
            {
            ($punct, $word, $data) = ($`, $&, $');
            if ($punct)
                { push(@{$$tags[$#$words + 1]},
                       new M1token('data:text', $punct)); }
            push(@$words, $word);
            $word =~ s/'//gi;
            next if $word eq '';

            my $pfrag;
            foreach $pfrag (@$pfrags)
                { $pfrag .= "-$word"; }          # append to each pfragment
            push(@$pfrags, $word);               # also start a new pfragment

            my $pfi;
            for ($pfi = 0; $pfi < @$pfrags; $pfi++)
                {
                $pfrag = $$pfrags[$pfi];
                if ($$fine{$pfrag})
                    {
                    &M1term::dbg("  link found on $pfrag");
                    delete $self->{'sought'}{$pfrag};

                    my $link;
                    foreach $link (@{$$fine{$pfrag}})
                        {
                        my ($preback, $postback) = $link->{'target'}->words();

                        my $preidx = @$words;    # count back $preback words
                        while ($preback--)       # but skip punct-only words
                            {
                            do { $preidx--; }
                            while ($$words[$preidx] !~ /[0-9a-z]/i);
                            }

                        my $postidx = @$words;   # count back $postback words
                        while ($postback--)      # but skip punct-only words
                            {
                            do { $postidx--; }
                            while ($$words[$postidx-1] !~ /[0-9a-z]/i);
                            }

                        $link->{'ident'} = $M1finder::linkid++;

                        my $linkstart = new M1token('link:start', $link);
                        my $linkend = new M1token('link:end', $link);

                                                 # cuddle text tightly
                        push(@{$$tags[$preidx]}, $linkstart);
                        unshift(@{$$tags[$postidx]}, $linkend);
                        }
                    }

                if ($$partial{$pfrag})           # keep this pfragment
                    { 
                    $keep = 1;
                    }
                else
                    {
                    my $extra = @$words - ($pfrag =~ tr/-/-/) - 1;
                    if ($extra && ! $keep) { $discard = $extra; }
                    splice(@$pfrags, $pfi--, 1); # discard this fragment
                    }
                }
            }
        if ($data)
            {
            my $index = $#$words + 1;
            push(@{$$tags[$index]}, new M1token('data:text', $data));
            }

        if (!@$pfrags)                           # okay to flush everything
            {
            $self->flush();
            }
        elsif ($discard)                         # flush as many as possible
            {
            $self->dequeue($discard);
            }
        }

    # not status, error, or data: must be a tag

    else
        {
        my $index = $#$words + 1;      # tag goes one spot after last word

        if ($type eq 'tag:a' and ($data->{'name'} or $data->{'id'}))
            {
            my $name = $data->{'name'} || $data->{'id'};
            my $links = $self->{'named'}->{$name};
            my $link;
            my $linkdisplay = (exists($data->{'linkdisplay'}) ||
                               $data->{'class'} =~ /\blinkdisplay\b/i);
            delete $self->{'end-named'};
            if ($linkdisplay)
                { push(@{$$tags[$index]}, new M1token('status:dispstart',
                       {'title' => $data->{'title'}, 'name' => $name})); }

            foreach $link (@$links)
                {
                push(@{$$tags[$index]}, new M1token('link:start', $link));
                unshift(@{$self->{'end-named'}},
                        new M1token('link:end', $link));
                }

            if ($linkdisplay)
                {
                push(@{$$tags[$index]}, new M1token('status:dispend'));
                delete $self->{'end-named'};
                }
            delete $self->{'named'}->{$name};
            }
        elsif ($type eq 'tag:a')
            { delete $self->{'end-named'}; }

        push(@{$$tags[$index]}, $token);

        if ($type eq 'tag:/a' and $self->{'end-named'})
            {
            push(@{$$tags[$index]}, @{$self->{'end-named'}});
            delete $self->{'end-named'};
            }
        }
    }

sub readlinks                          # get links from the link database
    {
    my ($self, $url) = @_;             # M1url: URL of target document
    my (@links) = $self->{'db'}->backlinks($url);

    my $link;
    foreach $link (@links) { $self->addlink($link); }
    if ($self->{'select'})
        {
        my $selurl = $url->relative('#' . $self->{'select'});
        $self->addlink(new M1link(0, $selurl, '', '', '', 'select'));
        }
    }

sub addlink                            # prepare to look for one link
    {
    my ($self, $link) = @_;

    &M1term::msg("addlink " . &repr::repr($link));

    if ($link->{'target'} and $link->{'target'}->{'frag'}) # fine-grained
        {
        my ($preidx, $postidx, @words) = $link->{'target'}->words();
        my $frag = $link->{'target'}->{'frag'};
        my $key = join('-', @words);
        $self->{'sought'}->{$key} = 1;

        if ($frag =~ /^'/ || $frag =~ /^'?.*\(.*\).*$/)    # back compat
            {
            if ($link->{'selected'} and $self->{'fine'}{$key})
                {
                ${$self->{'fine'}{$key}}[0]->{'selected'} = 1;
                }
            else
                {
                my ($word, $part);
                pop(@words);
                foreach $word (@words)
                    {
                    $part .= $word;
                    $self->{'partial'}{$part} = 1;         # mark partials
                    $part .= '-';
                    }
                push(@{$self->{'fine'}{$key}}, $link);     # store the link
                }
            }
        else                           # old-style named fragment
            {
            my $named = $self->{'named'};
            push(@{$$named{$frag}}, $link);
            }
        }
    elsif ($link->{'source'})                              # coarse-grained
        {
        push(@{$self->{'coarse'}}, $link);                 # store the link
        }
    }

sub dequeue                            # send off the first $num words/tags
    {
    my ($self, $num) = @_;
    my ($words, $tags) = ($self->{'words'}, $self->{'tags'});
    my $i;

    foreach $i (0..$#$tags)
        {
        next unless $$tags[$i] and (0+@{$$tags[$i]} == 1);
        my $tag = ${$$tags[$i]}[0];
        next unless $tag->{'type'} eq 'data:text';
        $$words[$i] = $tag->{'data'} . $$words[$i];
        $$tags[$i] = undef;
        }

    while ($num and (@$words or @$tags))
        {
        if ($$tags[0])                 # send any preceding tags
            {
            my $token;
            foreach $token (@{$$tags[0]}) { $self->send($token); }
            $$tags[0] = undef;
            }

        my $i;
        for ($i = 0; !$$tags[$i]; )    # skip to next tag in the queue
            { last if ++$i >= $num; }

        if ($i)                        # send the words up to that point
            {
            $self->send(new M1token('data:text', join('', @$words[0..$i-1])));
            my @foo = @$words; splice(@foo, 0, $i); @$words = @foo;
            my @foo = @$tags;  splice(@foo, 0, $i); @$tags  = @foo;
            $num -= $i;
            }
        }
    }

sub flush                              # flush everything from the queues
    {
    my ($self) = @_;
    while (@{$self->{'words'}} or @{$self->{'tags'}}) { $self->dequeue(100); }
    }

sub print                              # display the state for debugging
    {
    my ($self) = @_;
    my ($words, $tags) = @{$self}{'words', 'tags'};

    &M1term::dbg("    /----");
    &M1term::dbg("    finder queue: tags " . @$tags . ", words " . @$words);

    my $i;
    for ($i=0; $i <= $#$words; $i++)
        {
        if ($$tags[$i])
            {
            &M1term::dbg("      tags[$i] (".scalar(@{$$tags[$i]}).")");
            my $tag;
            foreach $tag (@{$$tags[$i]})
                { &M1term::dbg("        " . $tag->repr()); }
            }
        &M1term::dbg("      words[$i] = " . $$words[$i]);
        }
    &M1term::dbg("    ----/");
    }

1;
