# CritLink Mediator: Link-gathering package.
# Implements the M1gather class.  This sprocket looks for links in the
# incoming stream and stores them to the given link database.
#
# Exports: ->recv($token)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1gather;
use M1http;
use M1link;
use strict;

my $CONTEXT = 3;                       # words to keep before/after a link
my $HTTPSUCCESS = 200;                 # HTTP/1.0 response code for success
my $HTTPERROR = 400;                   # lower bound on codes for errors

sub new
    {
    my ($class, $db) = @_;
    return bless({'db'    => $db,      # M1db: link database to store in

                  'state' => 'resp',   # private: state
                  'words' => [],       # private: word queue
                  'tags'  => [],       # private: anchor tag queue
                  'links' => []},      # private: links to store
                 $class);
    }

sub recv
    {
    my ($self, $token) = @_;           # M1token: next token to examine
    my ($words, $tags) = @{$self}{'words', 'tags'};
    my ($type, $data) = @{$token}{'type', 'data'};

    if ($type eq 'status:url')
        {
        $self->{'base'} = $self->{'url'} = $data;
        $self->{'source'} = $self->{'db'}->docinfo($data);
        return;
        }

    if ($type =~ /^response:/)
        {
        if ($' == $HTTPSUCCESS)
            {
            my $info = $self->{'source'};
            delete $info->{'title'};
            delete $info->{'author'};
            delete $info->{'contributors'};
            delete $info->{'subject'};
            delete $info->{'keywords'};
            delete $info->{'notify'};
            $self->{'fetchok'} = 1;
            }
        if (($' >= $HTTPERROR) && $self->{'url'})
            {
            $self->{'db'}->store($self->{'url'});
            delete $self->{'source'}->{'last-modified'};
            &M1term::msg("clear links for " . $self->{'url'}->canon());
            }
        $self->{'state'} = 'out';
        return;
        }
    
    if ($self->{'state'} eq 'resp') { return; }
    if ($type eq 'status:title') { $self->{'source'}->{'title'} = $data; }

    if ($type eq 'header:last-modified')
        { $self->{'source'}->{'last-modified'} = $data; }

    if ($type eq 'status:start' and $self->{'fetchok'})
        {
        my $info = $self->{'source'};
        if (! $info->{'last-modified'})
            { $info->{'last-modified'} = &M1http::datetime(); }
        if (! $info->{'first-seen'})
            { $info->{'first-seen'} = $info->{'last-modified'}; }
        }

    # support for common metainformation names (including Dublin Core)

    if ($type eq 'tag:meta' and $$data{'content'})
        {
        my $info = $self->{'source'};
        if ($$data{'name'} =~ /^\s*(DC\.)?title\s*$/i)
            { $info->{'title'} = $$data{'content'}; }
        if ($$data{'name'} =~ /^\s*(DC\.)?(author|creator)\s*$/i)
            { $info->{'author'} = $$data{'content'}; }
        if ($$data{'name'} =~ /^\s*(DC\.)?contributors?\s*$/i)
            { $info->{'contributors'} .= $$data{'content'} . ', '; }
        if ($$data{'name'} =~ /^\s*(DC\.)?(subject|description)\s*$/i)
            { $info->{'subject'} = $$data{'content'}; }
        if ($$data{'name'} =~ /^\s*(DC\.)?keywords?\s*$/i)
            { $info->{'keywords'} = $$data{'content'}; }

        if ($$data{'name'} =~ /^\s*linknotify\s*$/i)
            { $info->{'notify'} = $$data{'content'} . '; '; }
        if ($$data{'name'} =~ /^\s*trigger\s*$/i)
            { $info->{'trigger'} = $$data{'content'}; }
        }

# coarse-grained tail

    if ($type eq 'tag:link')
        {
        my $target = $$data{'href'};
        my $relrole = $$data{'rel'};
        my $revrole = $$data{'rev'};
        my $title = $$data{'title'};
        $relrole =~ tr/A-Z/a-z/;
        $revrole =~ tr/A-Z/a-z/;

        if ($revrole eq 'made' or $relrole eq 'author')
            { 
            $self->{'source'}->{'author'} = $$data{'title'} or
            ($self->{'source'}->{'author'} = $$data{'href'}) =~ s/^mailto://;
            $self->{'source'}->{'authorlink'} = $$data{'href'};
            }

        if ($self->{'base'} and $target !~ /^#/ and
            ($target !~ /^(\w+):/ or $target =~ /^http:/i))
            {
            my $link = new M1link($self->{'url'},
                                  $self->{'base'}->relative($target),
                                  $title, $relrole, $revrole);
            &M1term::msg($link);
            push(@{$self->{'links'}}, $link);
            }

        return;
        }

    if ($type eq 'tag:base' and $$data{'href'})
        { $self->{'base'} = $self->{'base'}->relative($$data{'href'}); }

    if ($type eq 'error:eof' and $self->{'fetchok'})
        {
        if (exists($self->{'source'}->{'contributors'}))
            { $self->{'source'}->{'contributors'} =~ s/,\s*$//; }
        $self->{'db'}->store($self->{'url'}, @{$self->{'links'}});
        }

    if ($type eq 'tag:title') { $self->{'in-title'} = 1; }
    if ($type eq 'tag:/title') { $self->{'in-title'} = 0; }

    if ($type !~ /^(data:text|tag:\/?a|status:end)$/ or $self->{'in-title'})
        { return; }

    if ($type eq 'data:text')
        {
        $data =~ s/'//g;
        $data =~ s/[^0-9a-zA-Z]+/ /g;
        push(@$words, split(' ', $data));
        $#$tags = $#$words if $#$tags < $#$words;
        }
    elsif ($type eq 'tag:a' and $$data{'href'})
        {
        $$tags[$#$words + 1] = 'a ' .
            &M1med::unmediate($$data{'href'}) . ' ' . $$data{'rev'}; 
        $self->{'state'} = 'in';
        }
    elsif ($type eq 'tag:/a')
        {
        $$tags[$#$words + 1] = '/a';
        }

    if ($self->{'state'} eq 'out')
        {
        my $extra = @$words - $CONTEXT;
        if ($extra)
            {                          # Perl complains these are "read-only"
            my @foo = @$words; splice(@foo, 0, $extra); @$words = @foo;
            my @foo = @$tags;  splice(@foo, 0, $extra); @$tags = @foo;
            }
        return;
        }

    while (1)
        {
        my($i, $si, $ei, $target, $linktype);
        
        for ($i=0; $i < @$tags; $i++)            # find start anchor
            { last if $$tags[$i] =~ /^a /; }
        last if $$tags[$i] !~ /^a /;
        ($target, $linktype) = split(' ', $');
        $si = $i;

        for ($i++; $i < @$tags; $i++)            # find end anchor
            { last if $$tags[$i]; }
        if (!$$tags[$i]) { return; }
        $ei = $i;

        if ($ei + $CONTEXT > @$words and $type ne 'status:end')
            { return; }

        my $left = $si - $CONTEXT;
        my $right = $ei + $CONTEXT - 1;
        $left = 0 if $left < 0;
        $right = $#$words if $right > $#$words;

        my $pre = join('-', @$words[$left .. $si - 1]);
        my $anchor = '(' . join('-', @$words[$si .. $ei - 1]) . ')';
        my $post = join('-', @$words[$ei .. $right]);
        $anchor = "$pre-$anchor" if $pre;
        $anchor = "$anchor-$post" if $post;

# fine-grained tail

        if ($self->{'base'} and $target !~ /^#/ and
            ($target !~ /^(\w+):/ or $target =~ /^http:/i))
            {
            my $link = new M1link($self->{'url'}->relative("#:words:$anchor"),
                                  $self->{'base'}->relative($target),
                                  '',
                                  '',
                                  $linktype);
            &M1term::msg($link);
            push(@{$self->{'links'}}, $link);
            }

        $$tags[$si] = '';
        my $extra = $ei - $CONTEXT;
        if ($extra > 0)                # flush as many words/tags as possible
            {
            #splice(@$words, 0, $extra);
            #splice(@$tags, 0, $extra);
            @$words = @$words[$extra..$#$words];
            @$tags = @$tags[$extra..$#$tags];
            }
        }

    $self->{'state'} = 'out';
    }

1;
