# CritLink Mediator: HTTP package.
# Implements the M1http class.  This token-stream source opens
# an HTTP connection as a client and generates a stream of tokens
# from the incoming data.  It parses only the HTTP headers, but
# not the body of the retrieved entity.
#
# Exports: ->fetch($method, $url, $entity, @headers)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1http;
use M1config;
use M1token;
use M1term;
use repr;
use FileHandle;
use Socket;
use strict;

@M1http::ISA = 'sprocket';             # inherit from "sprocket"
$M1http::handle = 'aaaaaa';

# Regular expressions for the three date formats accepted in HTTP.
my $RFC1123_RE = q/([A-Z]\w\w), (\d+) ([A-Z]\w\w) (\d+) (\d+):(\d+):(\d+)/;
my $RFC850_RE  = q/([A-Z]\w+), (\d+)-([A-Z]\w\w)-(\d+) (\d+):(\d+):(\d+)/;
my $ASCTIME_RE = q/([A-Z]\w\w) ([A-Z]\w\w)\s+(\d+) (\d+):(\d+):(\d+) (\d+)/;

# Mapping between day or month names and numbers.
my %MONTHNUM = ('jan' => 0, 'january' => 0,
                'feb' => 1, 'february' => 1,
                'mar' => 2, 'march' => 2,
                'apr' => 3, 'april' => 3,
                'may' => 4,
                'jun' => 5, 'june' => 5,
                'jul' => 6, 'july' => 6,
                'aug' => 7, 'august' => 7,
                'sep' => 8, 'september' => 8,
                'oct' => 9, 'october' => 9,
                'nov' => 10, 'november' => 10,
                'dec' => 11, 'december' => 11);

my %DAYNUM =   ('sun' => 0, 'sunday' => 0,
                'mon' => 1, 'monday' => 1,
                'tue' => 2, 'tuesday' => 2,
                'wed' => 3, 'wednesday' => 3,
                'thu' => 4, 'thursday' => 4,
                'fri' => 5, 'friday' => 5,
                'sat' => 6, 'saturday' => 6);
            
my @MONTHNAME = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
my @DAYNAME = qw(Sun Mon Tue Wed Thu Fri Sat);

sub DESTROY
    {
    my ($self) = @_;
no strict 'refs';
    close($self->{'socket'});          # clean up
use strict 'refs';
    }

sub fetch
    {
    my ($self,
        $method,                       # string: HTTP request method
        $url,                          # M1url: URL of document to fetch
        $entity,                       # string: entity to POST, if any
        @headers) = @_;                # strings: any other headers to send

    $self->send(new M1token('status:url', $url));

    my ($server, $port) = @{$url}{'server', 'port'};
    my $path = $url->request();

    my ($name, $alias, $type, $len, $addr);
    my $proto = (getprotobyname('tcp'))[2];

    &M1term::msg("looking up $server");
    if ($server =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) # numeric IP address
        { $addr = pack('C4', $1, $2, $3, $4); }
    else
        { ($name, $alias, $type, $len, $addr) = gethostbyname($server); }

    if (!$addr)
        { $self->send(new M1token('error:hostname',
            $! || "Could not resolve hostname $server.")); return; }

    my $here = pack('S n a4 x8', AF_INET, 0, $M1config::localip);
    my $there = pack('S n a4 x8', AF_INET, $port, $addr);

    my $socket = $self->{'socket'} = $M1http::handle;
    $M1http::handle++;

no strict 'refs';                      # filehandle name is a symbolic ref

    socket($socket, PF_INET, SOCK_STREAM, $proto) || do
        { $self->send(new M1token('error:socket', $!)); return; };

    bind($socket, $here) || do
        { $self->send(new M1token('error:bind', $!)); return; };

    &M1term::msg("trying ".join('.',unpack('C4', $addr))." port $port");

    connect($socket, $there) || do
        { $self->send(new M1token('error:connect', $!)); return; };

    $socket->autoflush(1);

use strict 'refs';

    &M1term::msg("requesting $path");
    $socket->print("$method $path HTTP/1.0\r\n");
    $socket->print("User-Agent: $M1config::program\r\n");
    my $header;
    foreach $header (@headers)
        {
        if ($header =~ /user-agent:\s*(.*)/i)
            { &M1config::setclient($1); }
        else
            { $socket->print($header . "\r\n"); }
        if ($header =~ /cookie:\s*(.*)/i)
            { $self->send(new M1token("status:cookie", $1)); }
        }
    $socket->print("\r\n");
    $socket->print($entity);           # might be nothing

    my $data = '';

    # read response code

    while ($data !~ /^(.*)\n/)         # neet to get one whole line
        { if (!&read($data)) { return; } }

    $data =~ /^(.*)\n/;
    my $response = $&;
    $data = $';
    if ($response =~ m#^HTTP/\d+\.\d+\s+(\d+)\s*([^\r\n]*)#)
        { $self->send(new M1token("response:$1", $2)); }

    # read response headers

    my $datatype = 'data:binary';
    my $header;
    do
        {
        while ($data !~ /^(.*)\n/)     # need to get one whole line
            { if (!&read($data)) { return; } }

        $data =~ /^(.*)\n/;
        $header = $&;
        $data = $';
        $header =~ s/[\r\n]*$//;

        if ($header =~ /:\s*/)
            {
            my ($name, $value) = ($`, $');
            $name =~ tr/A-Z/a-z/;
            if ($name eq 'content-type' and $value =~ m#^text/#)
                { $datatype = 'data:text'; }
            if ($name eq 'content-location')
                { $url = $url->relative($value); }
            if ($name eq 'last-modified' or $name eq 'expires')
                {
                my ($s, $m, $h, $d, $l, $y, $w);
                (($w, $d, $l, $y, $h, $m, $s) = ($value =~ /$RFC1123_RE/)) ||
                    (($w, $d, $l, $y, $h, $m, $s) = ($value =~ /$RFC850_RE/)) ||
                    (($w, $l, $d, $h, $m, $s, $y) = ($value =~ /$ASCTIME_RE/));
                $w =~ tr/A-Z/a-z/; $w = $DAYNUM{$w};
                $l =~ tr/A-Z/a-z/; $l = $MONTHNUM{$l};
                if ($y < 1900) { $y += 1900; }
                $value = sprintf("%3s, %02d %3s %04d %02d:%02d:%02d GMT",
                    $DAYNAME[$w], $d, $MONTHNAME[$l], $y, $h, $m, $s);
                }
            $self->send(new M1token("header:$name", $value));
            }
        }
    while $header ne '';               # blank line means end of headers

    # read entity data

    while (1)
        {
        $self->send(new M1token($datatype, $data));
        $data = '';
        if (!&read($data)) { return; }
        }

    sub read                           # read some data from the socket
        {
        my ($rfd, $wfd, $efd) = ('', '', '');
no strict 'refs';
        my $sockfd = fileno($socket);
use strict 'refs';
        vec($rfd, $sockfd, 1) = 1;
        vec($efd, $sockfd, 1) = 1;

        my ($found, $left) = select($rfd, $wfd, $efd, $M1config::clienttimeout);
        if (!vec($rfd, $sockfd, 1)) { return &error('timeout'); }

        my $chunk;
no strict 'refs';
        my $bytes = sysread($socket, $chunk, 4096);
use strict 'refs';
        $_[0] .= $chunk;

        if ($bytes < 0) { return &error('read'); }
        if ($bytes == 0) { return &error('eof'); }
        &M1term::dbg(&repr::repr($chunk));
        return 1;
        }

    sub error
        {
        $self->send(new M1token('error:' . $_[0], $!));
no strict 'refs';
        close($socket);
use strict 'refs';
        return 0;
        }
    }

sub datetime
    {
    my ($s, $m, $h, $d, $l, $y, $w) = gmtime($_[0] or time());
    return sprintf "%3s, %2d %3s %4d %02d:%02d:%02d GMT",
        (qw/Sun Mon Tue Wed Thu Fri Sat/)[$w], $d,
        (qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/)[$l],
        $y+1900, $h, $m, $s;
    }

1;
