# CritLink Mediator: Lexical analysis package.
# Implements the M1lex class.  This sprocket receives tokens of
# textual data and parses out SGML tags as separate tokens in the
# outgoing stream.
#
# Exports: ->recv($token)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1lex;
use sgmlent;
use strict;

@M1lex::ISA = 'sprocket';              # inherit from "sprocket"

# SGML entities for special characters.
my %SGML_ENTITY = %sgmlent::SGML_ENTITY;

# Regular expression for an SGML declaration tag.
# The "---+" is not correct, but accommodates long strings of hyphens.
my $DECLTAG_RE     = q/<!                        # declaration tag-open
                       ([^>-]|-[^>-])*           # declaration text
                       (--([^-]|-[^-]|----+)*    # possible comment
                        (---+(?=>)|--)           # proper or brain-dead end
                        ([^>-]|-[^>-])*)*        # more declaration text
                       -?>/;                     # tag-close

# Regular expression for an HTML start tag.
my $STARTTAG_RE    = q/<\s*(\w+)                 # tag-open and element name
                       (([^>'"]|                 # unquoted text
                         '([^']|'\w)*'|          # single-quoted text
                         "[^"]*")*)              # double-quoted text
                       >/;                       # tag-close

# Regular expression for an HTML end tag.
my $ENDTAG_RE    =   q{<\s*/\s*(\w+)             # tag-open and element name
                       [^>]*                     # any (invalid) stray text
                       >};                       # tag-close

# Regular expression to find the end of a CDATA section,
# to be generated at runtime by editing the ENDTAG_RE constant.
my $ENDCDATA_RE;

# Regular expression for other invalid text that looks like a tag.
my $GARBAGE_RE    =  q/<\s*[^\s\w!]              # not a start, end, or decl
                       [^>]*                     # some text
                       >/;                       # possible tag-close

# Regular expression to extract the first of a list of attributes.
my $ATTRIB_RE    =   q/\s*([\w-]+)               # attribute name
                       (\s*=\s*                  # possible value
                       ('([^']*)'|               # single-quoted value
                        "([^"]*)"|               # double-quoted value
                        ([^\s>]+)))?/;           # unquoted value

# Remove the comments and whitespace from the above regex constants.
$DECLTAG_RE =~ s/\s*#.*\s*//g;
$STARTTAG_RE =~ s/\s*#.*\s*//g;
$ENDTAG_RE =~ s/\s*#.*\s*//g;
$GARBAGE_RE =~ s/\s*#.*\s*//g;
$ATTRIB_RE =~ s/\s*#.*\s*//g;

sub recv
    {
    my ($self, $token) = @_;

    if ($token->{'type'} eq 'header:content-type' and
        $token->{'data'} =~ m#^text/([^;\s]+)# and $1 ne 'html')
        { $self->{'plaintext'} = 'start';
        $token->{'data'} = 'text/html'; }
    
    if ($token->{'type'} eq 'status:url')
        { $self->{'url'} = $token->{'data'}; }

    if ($self->{'plaintext'})                    # wrap plain text in <PRE>
        {
        if ($self->{'plaintext'} eq 'start')
            {
            if ($token->{'type'} eq 'data:text')
                {
                my $filename = $self->{'url'}->canon('nofrag');
                $filename =~ s#.*/##;
                $self->{'queue'} = <<TEXTHEADER;
<!doctype html public "-//W3C//DTD HTML 3.2//EN">
<html><head>
  <title>$filename</title>
</head><body>
<h1>$filename</h1><pre>
TEXTHEADER
                $self->tokenize();
                $self->send(new M1token('data:text', $token->{'data'}));
                $self->{'plaintext'} = 'body';
                return;
                }
            }
        elsif ($self->{'plaintext'} eq 'body')
            {
            if ($token->{'type'} eq 'data:text')
                {
                $self->send(new M1token('data:text', $token->{'data'}));
                return;
                }
            elsif ($token->{'type'} eq 'error:eof')
                {
                $self->send(new M1token('tag:/pre'));
                $self->send(new M1token('tag:/body'));
                $self->send(new M1token('tag:/html'));
                return;
                }
            }
        }
    elsif ($token->{'type'} eq 'data:text')      # tokenize some HTML text
        {
        $self->{'queue'} .= $token->{'data'};
        $self->tokenize();
        return;
        }

    $self->tokenize();                       # tokenize anything left

    if ($self->{'queue'})                    # flush everything else
        { 
        $self->send(new M1token("data:binary", $self->{'queue'}));
        $self->{'queue'} = '';
        }

    $self->send($token);                     # send original token
    }

sub tokenize
    {
    my ($self) = @_;
    while ($self->{'queue'})
        {
        if ($self->{'cdata'})                    # ignore everything in CDATA
            {
            # the check for a double-quote is a hack to try to avoid
            # occurrences of </SCRIPT> inside JavaScript strings
            if ($self->{'queue'} =~ /$ENDCDATA_RE([^"])/i)
                {
                $self->{'queue'} = $1 . $';
                $self->send(new M1token('data:binary', $`));
                $self->send(new M1token('tag:/' . $self->{'cdata'}));
                $self->{'cdata'} = 0;
                }
            else { return; }
            }

        if ($self->{'queue'} =~ s/^$DECLTAG_RE//o)
            {                                    # tokenize declaration tags
            my $tag = $&;
            if ($tag =~ /---+/)
                {
                # Stupid!  Clean up the comment tag.
                $tag =~ s/[-!<>]//g;
                $tag = "<!-- Crit found a malformed comment: $tag -->";
                }
            $self->send(new M1token('tag:decl', $tag));
            }

        elsif ($self->{'queue'} =~ s/^$STARTTAG_RE//o)
            {                                    # tokenize start tags
            $self->sendtag($1, $2);
            }

        elsif ($self->{'queue'} =~ s/^$ENDTAG_RE//o)
            {                                    # tokenize end tags
            (my $tag = $1) =~ tr/A-Z/a-z/;
            $self->send(new M1token("tag:/$tag"));
            }

        elsif ($self->{'queue'} =~ s/^$GARBAGE_RE//o)
            {                                    # dispose of garbage
            $self->send(new M1token("data:binary", $&));
            }

        elsif ($self->{'queue'} =~ s/^([^<]+)</</)
            {                                    # match text content
            $self->sendtext($1);
            }

        elsif ($self->{'queue'} =~ s/^[^<]*\n//)
            {                                    # flush text after newline
            $self->sendtext($&);
            }

        else { return; }                         # nothing left
        }
    }
    
sub sendtag
    {
    my ($self, $tag, $rest, $space) = @_;
    $tag =~ tr/A-Z/a-z/;
    if ($tag eq 'script' or $tag eq 'style')     # start CDATA section
        {
        $self->{'cdata'} = $tag;
        ($ENDCDATA_RE = $ENDTAG_RE) =~ s#\(\\w\+\)#$tag#;
        }
    my %attr;
    while ($rest =~ s/$ATTRIB_RE//o)
        {
        my ($name, $value) = ($1, $4);
        $value = $5 if $value eq '';
        $value = $6 if $value eq '';
        $name =~ tr/A-Z/a-z/;
        $attr{$name} = $value;
        }
    $self->send(new M1token("tag:$tag", \%attr, $space));
    }

sub sendtext
    {
    my ($self, $text) = @_;
    $text =~ s/&(\w+);/$SGML_ENTITY{$1}/eg;
    $text =~ s/&#(\d+);/pack('C',$1)/eg;
    $self->send(new M1token('data:text', $text));
    }

1;
