# CritLink Mediator: Mediator package.
# Implements the M1med class.  This sprocket is responsible for
# rewriting URLs in a token stream so that the links may be
# followed through a mediator (pseudo-proxy).
#
# Exports: ->recv($token)

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1med;
use M1config;
use M1url;
use M1token;
use strict;

@M1med::ISA = 'sprocket';

# attributes of tags that contain URLs
my %ATTRIBS = ('a'      => 'href',
               'link'   => 'href',
               'area'   => 'href',
               'base'   => 'href',
               'form'   => 'action',
               'frame'  => 'src',
               'body'   => 'background',
               'img'    => 'src',
               'input'  => 'src',
               'object' => 'src',
               'applet' => 'codebase',
	       'script' => 'src');

# which tags we should mediate URLs for
my %MEDIATE;
for $_ (qw/ a link area base form frame /)
    { $MEDIATE{$_} = 1; }

sub recv
    {
    my ($self, $token) = @_;
    my ($type, $data) = @{$token}{'type', 'data'};

    if ($type eq 'status:url')                   # remember the base URL
        { $self->{'base'} = $data; }

    if ($type eq 'tag:base')                     # change the base URL
        {
        $self->{'base'} = $self->{'base'} ?
            $self->{'base'}->relative($$data{'href'}) : $$data{'href'};
        }

    if ($type eq 'tag:meta' and $$data{'http-equiv'} =~ /^\s*refresh\s*$/i)
        {
        $$data{'content'} =~ s/url\s*=\s*([^;]+)/
            'URL=' . $self->rewrite($1, 1)/ei;
        }

    if ($type =~ /^tag:/ and $ATTRIBS{$'} and !exists($$data{'nocrit'}))
        {
        my ($type, $attr) = ($', $ATTRIBS{$'});
        if ($type eq 'applet' and !$$data{$attr}) { $attr = 'code'; }
        $token = $self->rewritetag($attr, $type, $data) || $token;
        if ($M1config::clicaps{'javascript'} and
            $type eq 'a' and !$data->{'onmouseover'} and $data->{'href'})
            {
            $data->{'href'} =~ s/[\r\n]//g;
            (my $jtitle = $data->{'href'}) =~ s/[%'"]/'%'.unpack('H2', $&)/eg;
            $token->{'data'}->{'onmouseover'} =
                "window.status=unescape('$jtitle'); return true";
            $token->{'data'}->{'onmouseout'} =
                "window.status=''; return true";
            }
        }

    $self->send($token);
    }

sub rewritetag                         # fix a given attribute of a given tag
    {
    my ($self, $attr, $type, $data) = @_;

    if (!$data or !ref($data) or !$$data{$attr}) { return 0; }
    my %newdata = %$data;                        # copy the attribute data
    $newdata{$attr} = $self->rewrite($$data{$attr}, $MEDIATE{$type});
    return new M1token("tag:$type", \%newdata);
    }

sub rewrite                            # fix a single URL
    {
    my ($self, $url, $mediate) = @_;
    my $base = $self->{'base'};

    $url =~ s/[\r\n]//g;                         # some URLs wrap and break
    my $abs = $base->relative($url);             # find the absolute URL
    my $ext = ($abs->canon() =~ /.*\./ and $');  # find the file extension

    if ($mediate and $abs->{'scheme'} eq 'http' and !$M1config::direct{$ext})
        { &M1term::dbg("mediating $url with extension $ext");
        return &mediate($abs->canon()); }
    else
        { return $abs->canon(); }
    }

# static methods for converting to/from mediated URLs
sub mediate
    {
    my $url = &unmediate($_[0]);
    my ($query, $frag);

    # workaround for MSIE 4 bug which replaces // with /
    $url =~ s#(ftp|http):/+#$1://#g;

    if ($url =~ /#/)
        { ($url, $frag) = ($`, $'); }
    if ($url =~ /\?/)
        { ($url, $query) = ($`, $'); }
    $frag =~ s/^([-\w]*\([-\w]*\)[-\w]*)$/'$1/;
    $url .= "%ff$frag" if $frag ne '';
    $url .= "?$query" if $query ne '';
    $url .= "#M1-target" if $frag ne '';

    return $M1config::medurl . '/' . $url;
    }

# initialize $medurlmatcher as a static for &unmediate
    {
    my ($medurlmatcher) = $M1config::medurl;
    $medurlmatcher =~ s#^/*##;
    $medurlmatcher =~ s#/*$##;
    $medurlmatcher =~ s#\W#\\$&#g;
    $medurlmatcher =~ s#/+#/+#g;

    sub unmediate
        {
        my ($url) = $_[0];
        my ($frag, $query);

        $url =~ s#^\s*|\s*$##g;
        $url =~ s#(ftp|http):/+#$1://#g;         # workaround for MSIE 4 bug
                                                 # which replaces // with /
        if ($url =~ /#/)
            { $url = $`; }
        if ($url =~ /\?/)
            { ($url, $query) = ($`, $'); }
        if ($url =~ /\xff|%ff/i)
            { ($url, $frag) = ($`, $'); }

        $url = &M1url::decode($url);

        # remove any copies of the URL prefix as long as another URL follows
        if ($url =~ m#^(/*($medurlmatcher/*)*)(\w+:/)# && $1 ne '')
            { $url = $3 . $'; }
        else
            { return $_[0]; }                    # not mediated; leave it alone
        $url = &M1url::encode($url);

        $frag =~ s/^([-\w]*\([-\w]*\)[-\w]*)$/:words:$1/;
        $url .= "?$query" if $query ne '';
        $url .= "#$frag" if $frag ne '';
        return $url;
        }
    }

1;
