# CritLink Mediator: User interface package.
# Implements the M1ui class.  This sprocket is responsible for
# dealing with link and start/end indicators in the token stream
# and replacing them with corresponding user interface components.
#
# Exports: ->recv()

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1ui;
use M1url;
use M1med;
use strict;

@M1ui::ISA = 'sprocket';               # inherit from "sprocket"
my $HEADER_TEMPLATE = "$M1config::docpath/cl-header.html";
my $PREFOOTER_TEMPLATE = "$M1config::docpath/cl-prefooter.html";
my $POSTFOOTER_TEMPLATE = "$M1config::docpath/cl-postfooter.html";
my $PRELINKLIST_TEMPLATE = "$M1config::docpath/cl-prelinklist.html";
my $POSTLINKLIST_TEMPLATE = "$M1config::docpath/cl-postlinklist.html";
my $MORELINKS_TEMPLATE = "$M1config::docpath/cl-morelinks.html";
my $PRELINKDISP_TEMPLATE = "$M1config::docpath/cl-prelinkdisp.html";
my $POSTLINKDISP_TEMPLATE = "$M1config::docpath/cl-postlinkdisp.html";
my $ERROR_TEMPLATE = "$M1config::docpath/cl-error.html";
my $ANCHOR_TEMPLATE = "$M1config::docpath/cl-anchor.html";

my %LINK_COLOUR = ('query'   => 'o',   # orange
                   'comment' => 'b',   # blue
                   'support' => 'g',   # green
                   'issue'   => 'r');  # red
sub linkicon
    {
    my ($side, $link) = @_;
    my $colour = $LINK_COLOUR{$link->{'rev'}} || 'p';
    if ($link->{'selected'})
        { return ('src' => "$M1config::iconurl/sel-$side-$colour.gif",
                  'width' => 11, 'height' => 9); }
    else
        { return ('src' => "$M1config::iconurl/link-$side-$colour.gif",
                  'width' => 9, 'height' => 9); }
    }

my %LINK_SYMBOL = ('query'   => '?',
                   'comment' => '#',
                   'support' => '+',
                   'issue'   => '-');
sub textlinkicon
    {
    my ($side, $link) = @_;
    my $symbol = $LINK_SYMBOL{$link->{'rev'}} || '';
    if ($side eq 'l')
        {
        return $link->{'selected'} ? ">$symbol>>" : "[$symbol>";
        }
    else
        {
        return $link->{'selected'} ? "<<$symbol<" : "<$symbol]";
        }
    }

sub new
    {
    my ($class, $next, $db, $tfrag) = @_;
    return bless({'next'    => $next,  # sprocket: the next sprocket
                  'db'      => $db,    # M1db: database to read from
                  'tfrag'   => $tfrag, # string: requested target fragment

                  'coarse'  => [],     # private: list of incoming coarse links
                  'links'   => [],     # private: list of outgoing coarse links
                  'missing' => [],     # private: list of missing links
                  'found'   => 0},     # private: count of target phrases found
                 $class);
    }

sub recv
    {
    my ($self, $token) = @_;
    my ($type, $data) = @{$token}{'type', 'data'};
    &M1term::dbg($token);

    # local subroutines for constructing output

no strict 'vars';
    sub sendtext { $self->send(new M1token('data:text', $_[0])); }
    sub sendtag { $self->send(new M1token("tag:$_[0]", $_[1])); }
    sub sendraw { $self->send(new M1token('data:binary', $_[0])); }

    sub makeheader
        {
        local ($docurl) = $self->{'url'} && $self->{'url'}->canon('nofrag');
        local ($requrl) = $self->{'url'} && $self->{'url'}->canon();
        local ($docencurl) = &M1url::encode($docurl);
        local ($telleractive) = ($self->{'cookie'} =~ /teller-active/i);
        local ($editbutton, $commentbutton);
        if ($docurl =~ $M1config::puburlmatcher)
            {
            local($addrdomain, $addruser, $path) = ($1, $2, $3);
            if (-f "$M1config::pubpath/$addrdomain/$addruser/$path")
                { $editbutton = 1; }
            }
        if ($self->{'respcode'} == 200)
            { $commentbutton = 1; }
        return &render($HEADER_TEMPLATE, 'M1config');
        }

    sub sendanchor                     # for commenting at a named anchor
        {
        local ($targetanchor) = $_[0];
        local ($docurl) = $self->{'url'} && $self->{'url'}->canon('nofrag');
        local ($docencurl) = &M1url::encode($docurl);
        local ($commentbutton) = &render($ANCHOR_TEMPLATE, 'M1config');
        $commentbutton =~ s/^\s*|\s*$//g;
        sendraw $commentbutton;
        }

    sub senderror
        {
        local ($respcode, $respmsg, $title, $message) = @_;
        local ($docurl) = $self->{'url'} && $self->{'url'}->canon('nofrag');
        local ($requrl) = $self->{'url'} && $self->{'url'}->canon();
        local ($header);
        if (! $self->{'started'})
            {
            $self->send(new M1token("response:$respcode", $respmsg));
            $self->send(new M1token('header:content-type', 'text/html'));
            $header = &makeheader;
            }
        sendraw &render($ERROR_TEMPLATE);
        }

    sub sendbacklink
        {
        my ($link, @anchor) = @_;
        my ($rel, $rev, $source) = @{$link}{'rel', 'rev', 'source'};
        my $info = $self->{'db'}->docinfo($source);

        my $title;
        $title .= "[$rev] " if $rev;
        $title .= $info->{'author'} . ': ' if $info->{'author'};
        $title .= $info->{'title'};

        my $attr = {            $source ? ('href' => $source->canon()) : (),
                    $link->{'selected'} ? ('name' => 'M1-target')      : (),
                                   $rev ? ('rel' => $rev)              : (),
                                   $rel ? ('rev' => $rel)              : () };

        if ($title ne '')
            {
            (my $htitle = $title) =~ s/"/&quot;/g;
            $$attr{'title'} = $htitle;
            if ($M1config::clicaps{'javascript'})
                {
                (my $jtitle = $title) =~ s/[%'"]/'%'.unpack('H2', $&)/eg;
                $$attr{'onmouseover'} =
                    "window.status=unescape('$jtitle'); return true";
                $$attr{'onmouseout'} = "window.status=''; return true";
                }
            }

        if ($M1config::clicaps{'tooltips'})
            {
            foreach $tag (@anchor)
                {
                if ($tag->{'type'} eq 'tag:img')
                    { $tag->{'data'}->{'alt'} = $title;
                    delete($attr->{'onmouseover'}); }
                }
            }

        sendtag '/a' if $self->{'anchortag'};
        sendtag 'a', $attr;
        foreach $tag (@anchor)
            { $self->send($tag); }
        sendtag '/a';
        $self->send($self->{'anchortag'}) if $self->{'anchortag'};
        }

    sub sendauthor
        {
        my ($info) = $_[0];

        if ($info->{'author'})
            {
            my $link = $info->{'authorlink'};
            if (! $link and $info->{'author'} =~ /[\w.=-]+@[\w.-]+\.\w\w\w?/)
                { $link = "mailto:$&"; }
            sendtext ' by ';
            sendtag 'a', {'href' => $link, 'rel' => 'author'} if $link;
            sendtext $info->{'author'};
            sendtag '/a' if $link;
            }
        }

    sub sendcoarse
        {
        my ($titleurl, $fragurl, $rel, $rev) = @_;
        my ($link, $frag) = @_;
        my $info = $self->{'db'}->docinfo($titleurl);
        my $attr = { 'href' => $titleurl->canon('nofrag'),
                     $rev ? ('rel' => $rev) : (),
                     $rel ? ('rev' => $rel) : () };
        sendtag 'a', $attr;
        sendtag 'font', {'color' => '#007000'};
        sendtext ($info->{'title'} or $$attr{'href'});
        sendtag '/font';
        sendtag '/a';
        sendauthor($info);
        sendtext "\n";
        $$attr{'href'} = $titleurl->canon();
        sendfrag($fragurl, $attr);
        sendtext "\n";
        }

    sub sendfrag
        {
        my ($url, $attr) = @_;
        my ($preback, $postback, @words) = $url->words();
        if (@words)
            {
            my ($pre, $post) = (@words - $preback, @words - $postback);
            my $prewords = join(' ', @words[0..$pre-1]);
            $prewords = "$prewords " if $prewords;
            my $postwords = join(' ', @words[$post..$#words]);
            $postwords = " $postwords" if $postwords;

            sendtag 'small';
            sendtext " (...$prewords";
            sendtag 'a', $attr if $attr;
            sendtag 'strong';
            sendtag 'font', {'color' => '#007000'} if $attr;
            sendtext join(' ', @words[$pre..$post-1]);
            sendtag '/font' if $attr;
            sendtag '/strong';
            sendtag '/a' if $attr;
            sendtext "$postwords...)";
            sendtag '/small';
            }
        elsif ($url->{'frag'})
            {
            sendtag 'small';
            sendtext ' (';
            sendtag 'a', $attr if $attr;
            sendtag 'strong';
            sendtag 'font', {'color' => '#007000'} if $attr;
            sendtext '#' . $url->{'frag'};
            sendtag '/font' if $attr;
            sendtag '/strong';
            sendtag '/a' if $attr;
            sendtext ')';
            sendtag '/small';
            }
        }

    # ->recv() method starts here

    if ($type =~ /^response:/)
        { $self->{'respcode'} = $'; }

    if ($type eq 'error:nourl')
        {
        senderror 200, 'No URL Given', 'No URL given',
            'Please enter the URL of a document to view in the field above.';
        return;
        }
    if ($type eq 'error:badscheme')
        {
        senderror 503, 'Service Unavailable', 'Unsupported scheme',
            'Sorry, we only support the HTTP scheme for fetching documents.';
        return;
        }
    if ($type eq 'error:badmethod')
        {
        senderror 405, 'Method Not Allowed', 'Unknown method',
            "Sorry, we are unable to process the request method '$data'.";
        return;
        }
    if ($type ne 'error:eof' and $type =~ /^error:/)
        {
        senderror 500, 'Internal Server Error', "Error: $'",
            ($data ? "The system error message is '$data'." : '');
        return;
        }

    if ($type eq 'status:cookie')
        { $self->{'cookie'} = $data; }

    if ($type eq 'link:coarse')
        { push(@{$self->{'coarse'}}, $data); return; }

    if ($type eq 'link:missing')
        { push(@{$self->{'missing'}}, $data); return; }

    if ($type eq 'status:dispstart')
        { $self->{'linkdisplay'} = $data; }

    if ($type eq 'link:start' and exists($self->{'linkdisplay'}))
        { push(@{$self->{'linklist'}}, $data); return; }
    
    if ($type eq 'status:dispend')
        {
        my %info = %{$self->{'linkdisplay'}};
        sendraw &render($PRELINKDISP_TEMPLATE);

        sendtext 'Links related to ';
        sendtag 'strong';
        sendtext $info{'title'} || '(no topic)';
        sendtag '/strong';
        sendtext ':';
        if ($self->{'linklist'})
            {
            sendtag 'ol';
            my $link;
            for $link (@{$self->{'linklist'}})
                {
                sendtag 'li';
                if ($link{'rev'})
                    { my $art = ($attr{'rev'} =~ /^[aeiou]/i) ? 'an' : 'a';
                    sendraw "$art <em>$link{'rev'}</em> entitled "; }
                sendcoarse @{$link}{'source', 'source', 'rel', 'rev'};
                }
            sendtag '/ol';
            }
        else
            { sendraw '<p><em>(None yet.)</em> '; }

        local ($docurl) = $self->{'url'} && $self->{'url'}->canon('nofrag');
        sendtext 'To add to this list, create a document linked to ';
        sendtag 'tt';
        sendtext $self->{'url'}->relative('#' . $info{'name'})->canon();
        sendtag '/tt';
        sendraw &render($POSTLINKDISP_TEMPLATE);

        delete $self->{'linkdisplay'};
        delete $self->{'linklist'};
        }

    if ($type eq 'link:start')
        {
        sendbacklink $data, new M1token('tag:img',
                             {&linkicon('l', $data),
                              'align' => 'textbottom',
                              'border' => 0,
                              'alt' => &textlinkicon('l', $data)});
        $self->{'found'}++;
        return;
        }

    if ($type eq 'link:end')
        {
        sendbacklink $data, new M1token('tag:img',
                             {&linkicon('r', $data),
                              'align' => 'textbottom',
                              'border' => 0,
                              'alt' => &textlinkicon('r', $data)});
        return;
        }

    if ($type eq 'status:url')
        { $self->{'url'} = $self->{'base'} = $data; }

    if ($type eq 'tag:base' and $data->{'href'})
        { $self->{'base'} = $self->{'base'}->relative($data->{'href'}); }

    if ($type eq 'status:start')
        {
        if (! ($self->{'sent-header'}++)) { sendraw &makeheader; }
        $self->{'started'} = 1;
        }

    if ($type eq 'status:fulllisting')
        { $self->{'fulllisting'} = 1; }

    if ($type eq 'status:end')
        {
        my $info = $self->{'db'}->docinfo($self->{'url'});
        local $title = ($info->{'title'} || '(no title)');

        sendraw &render($self->{'fulllisting'} ? $PRELINKLIST_TEMPLATE :
                                                 $PREFOOTER_TEMPLATE);

        # Display information about the document.

        sendtag 'strong';
        sendtext $title;
        sendtag '/strong';
        sendauthor($info);
        if ($info->{'last-modified'} =~ /(\d+) ([A-Z]\w\w) (\d+)/)
            { sendtext ', last modified on ' . ($1+0) . " $2 $3" }
        sendtext '. ';

        my $key;
        foreach $key ('contributors', 'subject', 'keywords')
            {
            if ($info->{$key} ne '')
                {
                my $caps = $key;
                substr($caps, 0, 1) =~ tr/a-z/A-Z/;
                sendtext "$caps: ";
                sendtag 'small';
                $info->{$key} =~ s/[.,]?\s*$//;
                sendtext $info->{$key} . ".";
                sendtag '/small';
                sendtext ' ';
                }
            }

        # Display monitors (for now).
        if ($info->{'monitor'} =~ /\S/)
            {
            sendtext "Monitors: ";
            sendtag 'small';
            sendtext join(', ', split(' ', $info->{'monitor'})) . '.';
            sendtag '/small';
            sendtext ' ';
            }

        # Display information about the document's links.

        if (@{$self->{'links'}})
            {
            sendtext "This document is linked:\n";
            sendtag 'ol';
            my $link;
            foreach $link (@{$self->{'links'}})
                {
                &M1term::dbg("displaying link " . $link->repr());
                sendtag 'li';
                my $info = $self->{'db'}->docinfo($link->{'target'});
                my %attr = ('href' => $link->{'target'}->canon('nofrag'),
                            $link->{'rel'} ? ('rel' => $link->{'rel'}) : (),
                            $link->{'rev'} ? ('rev' => $link->{'rev'}) : ());

                if ($attr{'rev'})
                    {
                    if ($attr{'rev'} eq 'made')
                        { sendraw 'as <em>made</em> by '; }
                    else
                        {
                        my $art = ($attr{'rev'} =~ /^[aeiou]/i) ? 'an' : 'a';
                        sendraw "as $art <em>$attr{'rev'}</em> to ";
                        }
                    }
                elsif ($attr{'rel'})
                    {
                    sendraw "to a <em>$attr{'rel'}</em> document at ";
                    }
                else 
                    { sendtext 'to '; }
                sendtag 'a', \%attr;
                sendtag 'font', {'color' => '#007000'};
                sendtext $info->{'title'} ||
                         $link->{'title'} ||
                         $link->{'target'}->canon('nofrag');
                sendtag '/font';
                sendtag '/a';
                sendtext ' by ' . $info->{'author'} if $info->{'author'};
                $attr{'href'} = $link->{'target'}->canon();
                sendfrag($link->{'target'}, \%attr);
                sendtext "\n";
                }
            sendtag '/ol';
            $self->{'links'} = [];
            }

        my ($found, $missing, $coarse) =
            ($self->{'found'}, $self->{'missing'}, $self->{'coarse'});

        if ($self->{'found'})
            {
            sendtext "Fine-grained links ($self->{'found'}) are displayed\n";
            sendtag 'img', {&linkicon('l'),
                            'align' => 'textbottom',
                            'alt' => &textlinkicon('l')};
            sendtext 'like this';
            sendtag 'img', {&linkicon('r'),
                            'align' => 'textbottom',
                            'alt' => &textlinkicon('r')};
            sendtext "\nin the document text above.\n";
            }

        if (@$missing)
            {
            sendtext "Unable to locate targets for fine-grained links from:\n";
            sendtag 'ol';
            my $link;
            for $link (@$missing)
                {
                sendtag 'li';
                if ($link->{'source'})
                    { sendcoarse @{$link}{'source', 'target', 'rel', 'rev'}; }
                else
                    {
                    sendtext 'selected reference';
                    sendfrag $link->{'target'};
                    sendtext "\n";
                    }
                }
            sendtag '/ol';
            $self->{'missing'} = [];
            }

        if (@$coarse)
            {
            sendtext 'Coarse-grained links to this document exist from:';
            sendtag 'ol';
            my $link, $count;
            for $link (reverse @$coarse)
                {
                last if ++$count > 15 && !$self->{'fulllisting'};
                sendtag 'li';
                if ($$link{'rev'})
                    { my $art = ($$link{'rev'} =~ /^[aeiou]/i) ? 'an' : 'a';
                    sendraw "$art <em>$$link{'rev'}</em> entitled "; }
                sendcoarse @{$link}{'source', 'source', 'rel', 'rev'};
                }
            sendtag '/ol';
            if ($count > 15 && !$self->{'fulllisting'})
                {
                local $docurl = $self->{'url'}->canon();
                local $count = scalar(@$coarse);
                sendraw &render($MORELINKS_TEMPLATE);
                }
            $self->{'coarse'} = [];
            }
            
        local($docurl) = $self->{'url'} && $self->{'url'}->canon('nofrag');
        sendraw &render($self->{'fulllisting'} ? $POSTLINKLIST_TEMPLATE :
                                                 $POSTFOOTER_TEMPLATE);
        }

    if ($type eq 'tag:a')
        {
        $self->{'anchortag'} = $token; # save anchor tag in case interrupted
        if ($data->{'name'} =~ /\S+/)  # produce an "add comment" button
            { &sendanchor($&); }
        }
    if ($type eq 'tag:/a') { $self->{'anchortag'} = undef; }
    if ($type eq 'tag:link')
        {
        push(@{$self->{'links'}},
            new M1link($self->{'url'},
                       $self->{'base'}->relative($$data{'href'}),
                       $data->{'title'},
                       $data->{'rel'},
                       $data->{'rev'}));
        }

    if ($type eq 'tag:script') { $self->{'in-script'} = 1; }
    if ($type eq 'tag:/script') { $self->{'in-script'} = 0; }

    $self->send($token);
    }

# Routine for rendering a document from a template containing __fields__.
no strict;
{
my @packages;

sub render
    {
    my $file = shift;
    @packages = (@_, (caller)[0]) if @_;

    sub lookup
        {
        my ($symbol) = @_;
        my $package;
        if ($symbol =~ s/\[(\w+)\]//)
            {
            my $key = $1;
            foreach $package (@packages)
                {
                if (defined(%{$package.'::'.$symbol}))
                    { return ${$package.'::'.$symbol}{$key}; }
                }
            }
        else
            {
            foreach $package (@packages)
                {
                if (defined(${$package.'::'.$symbol}))
                    { return ${$package.'::'.$symbol}; }
                }
            }
        }

    sub renderbody
        {
        my $data = $_[0];
        my ($result, $command);
        while ($data =~ /__(([^_]|_[^_])+)__/)
            {
            $result .= $`;
            ($command, $data) = ($1, $');

            if ($command =~ /^\[(.*)\]$/)
                { $result .= &render("$M1config::docpath/$1"); }
            elsif ($command =~ s/^{// && $command =~ s/}$//)
                { $result .= eval($command); }
            else
                { $result .= &lookup($command); }
            }
        return $result . $data;
        }

    open(FILE, $file);
    my $data = join('', <FILE>);
    close(FILE);

    my ($result, $pre, $neg, $cond, $body);
    while ($data =~ /__\((!?)(\S+)\s(([^)]|\)[^_]|\)_[^_])*)\)__/)
        {
        ($pre, $neg, $cond, $body, $data) = ($`, $1, $2, $3, $');
        $result .= &renderbody($pre);
        $result .= &renderbody($body)
            if (&lookup($cond) ? 1 : 0) ^ ($neg ? 1 : 0);
        }
    $result .= &renderbody($data);
    return $result;
    }
}

1;
