# CritLink Mediator: URL-handling package.
# Implements the M1url class.  This object represents a URL.  It
# provides methods for reconstructing the URL in canonical form,
# for resolving a relative URL, and for converting to a filename.
#
# Exports: ->canon(), ->repr(), ->relative($urlstring),
#          ->request(), ->filename(), ->words()

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package M1url;
use M1config;
use strict;

# Regular expression for a filename which might match a trailing slash.
my $NULLNAME_RE = q/|(index|welcome|default)\.(s?html?|asp)/;

sub new
    {
    my ($class, $url) = @_;
    my $self = {};
    bless($self, $class);

    $url =~ s/^\s*|\s*$//g;
    if ($url =~ /#/)
        {
        ($url, $self->{'frag'}) = ($`, $');
        if ($self->{'frag'} =~ /^[-\w]*\([-\w]*\)[-\w]*$/)
            { $self->{'frag'} = "'" . $self->{'frag'}; }
        }
    if ($url =~ /\?/)
        { ($url, $self->{'query'}) = ($`, $'); }
    if ($url =~ /^([a-zA-Z]+):(\/*)/)                      # "scheme://" given
        {
        ($self->{'scheme'} = $1) =~ tr/A-Z/a-z/;
        if ($self->{'scheme'} ne 'http')
            { $self->{'opaque'} = $2 . $'; return $self; }

        ($url = $') =~ /^([^\/]*)/;
        $self->{'server'} = $&;
        $url = $';
        if ($self->{'server'} =~ /:/)
            { $self->{'server'} = $`;
            $self->{'port'} = $'; }
        }
    elsif ($url =~ /^([\w-]+\.[\.\w-]+)(:(\d+))?/)         # "http://" omitted
        {
        $self->{'scheme'} = 'http';
        $self->{'server'} = $1;
        $self->{'port'} = $3 if $2; 
        $url = $' || '/';
        }
    $self->{'server'} =~ tr/A-Z/a-z/;
    $self->{'server'} = &encode(&decode($self->{'server'}));
    if (! $self->{'port'})
        { $self->{'port'} = &defport($self->{'scheme'}); }
    $self->setpath($url);
    return $self;
    }

sub defport
    {
    my ($scheme) = @_;
    if ($scheme eq 'http') { return 80; }
    return 80;
    }

sub append
    {
    my ($self, $url) = @_;
    my $dir;
    for $dir (split('/', $url, -1))
        {
        if ($dir eq '.') { }
        elsif ($dir eq '..') { pop(@{$self->{'path'}}); }
        else { push(@{$self->{'path'}}, &encode(&decode($dir))); }
        }
    if (! @{$self->{'path'}})
        { @{$self->{'path'}} = (''); }
    }

sub setpath
    {
    my ($self, $path) = @_;            # string: a path name

    $self->{'path'} = [];
    $path =~ s#^/##;
    $self->append($path);
    }

sub relative
    {
    my ($self, $url) = @_;             # string: the relative URL
    $url =~ s/^\s*|\s*$//g;
    if ($url =~ /^$self->{'scheme'}:([^\/])/) { $url = $1 . $'; }
    elsif ($url =~ /^([a-zA-Z]+):\/*/) { return new M1url($url); }

    my $result = bless({'scheme' => $self->{'scheme'},
                        'server' => $self->{'server'},
                        'port'   => $self->{'port'}}, ref($self));
    if ($url =~ /#/) { ($url, $result->{'frag'}) = ($`, $'); }
    if ($url =~ /\?/) { ($url, $result->{'query'}) = ($`, $'); }

    my $dir;
    for $dir (@{$self->{'path'}})
        { push(@{$result->{'path'}}, $dir); }

    if ($url =~ /^\//)
        { $result->setpath($url); }
    else
        {
        pop(@{$result->{'path'}}) if $url;
        $result->append($url);
        }

    return $result;
    }

sub serverport
    {
    my ($self) = @_;

    if ($self->{'port'} != &defport($self->{'scheme'}))
        { return $self->{'server'} . ':' . $self->{'port'}; }
    else
        { return $self->{'server'}; }
    }

sub canon
    {
    my ($self, $options) = @_;

    if (! $self->{'scheme'})
        { return join('/', @{$self->{'path'}}); }

    if ($self->{'scheme'} ne 'http')
        { return $self->{'scheme'} . ':' . $self->{'opaque'}; }

    my $scheme = $self->{'scheme'} ? $self->{'scheme'}.'://' : '';
    my $serverport = $self->serverport();
    my $path = join('/', @{$self->{'path'}});
    my $query = $self->{'query'} ? '?' . $self->{'query'} : ''
        unless $options =~ /noquery/;
    my $frag = $self->{'frag'} ? '#' . $self->{'frag'} : ''
        unless $options =~ /nofrag/;
    $frag =~ s/^#'?([-\w]*\([-\w]*\)[-\w]*)$/#:words:$1/;
    return $scheme . $serverport . '/' . $path . $query . $frag;
    }

sub repr { return &canon; }

sub request
    {
    my ($self) = @_;
    my $request = '/' . join('/', @{$self->{'path'}});
    $request .= '?' . $self->{'query'} if $self->{'query'} ne '';
    return $request;
    }

sub filename
    {
    my ($self, $base) = @_;
    my $scheme = $self->{'scheme'};
    my @dirs = ($scheme);

    $base =~ s/\/*$//;
    if ($scheme =~ /^(http|https|ftp|telnet|news|finger|gopher)$/)
        {
        my $serverport = $self->serverport();
        my $port = ($serverport =~ s/:.*//) ? $& : '';
        my @chunks;

        if ($serverport =~ /^[0-9.]*$/)
            {
            # for IP numbers, just split in given order
            @chunks = split('\.', $serverport);
            }
        else
            {
            # for domain names, reorder components most to least significant
            @chunks = reverse split('\.', $serverport);

            if (@chunks >= 2)
                {
                # add an extra level before the second-last domain component
                splice(@chunks, 1, 0, substr($chunks[1], 0, 2));
                }
            }

        if ($port)
            {
            # tack the port number onto the least significant component
            $chunks[$#chunks] .= $port;
            }

        push(@dirs, @chunks);

        # add a delimiter to divide the host components from path components
        push(@dirs, '-');

        if ($scheme =~ /^(http|https|ftp)$/)
            { push(@dirs, @{$self->{'path'}}); }
        }
    else
        { push(@dirs, $self->{'opaque'}); }

    my $dir;
    for $dir (@dirs)
        {
        if (-f $base)
            {
            &M1term::dbg("moving $base to $base/\\");
            open(FILE, $base);
            my $data = join('', <FILE>);
            close(FILE);
            unlink($base);
            mkdir($base, 0755);
            open(FILE, "> $base/\\");
            print FILE $data;
            close(FILE);
            }
# Don't make parent dirs automatically -- generates too much random crud.
#        else
#            { mkdir($base, 0755); }
        $base .= "/$dir";
        }

    if (-d $base)
        { $base =~ s#/?$#/\\#; }

    if ($base =~ m#/($NULLNAME_RE)$#)
        { $base = $` . "/\\"; }

    $base .= '?' . $self->{'query'} if $self->{'query'};

    return $base;
    }

sub words                              # get the list of words in the
    {                                  # fragment and indices of its phrase
    my ($self) = @_;
    return undef unless $self->{'frag'} =~ /^(:words:|')?(.*)\((.*)\)(.*)/;

    my ($pre, $anchor, $post) = ($2, $3, $4);
    $pre =~ s/^-+//;
    $post =~ s/-+$//;
    $pre = "$pre-" if $pre =~ /[^-]$/;
    $post = "-$post" if $post =~ /^[^-]/;
    my @words = split(/-+/, "$pre$anchor$post");

    my $prelen = ($pre =~ tr/-/-/);    # words before (
    my $postidx = ($post =~ tr/-/-/);  # words after )

    return (scalar(@words) - $prelen, $postidx, @words);
    }

# static methods for %nn URL encoding
sub decode
    {
    my ($url) = @_;
    $url =~ s/%([0-9a-fA-F][0-9a-fA-F])/ pack('H2', $1) /eg;
    return $url;
    }

sub encode
    {
    my ($url) = @_;
    $url =~ s#[^\w._/*(),'\$\@:;&!~-]# '%' . unpack('H2', $&) #eg;
    return $url;
    }

1;
