# FTP PUT package.
# For storing files on an FTP server.

# The CritLink Mediator and CritWriter software is copyright
# 1997 by Ka-Ping Yee (http://www.lfw.org/ping/).  All rights
# reserved.  Written for a project of the Foresight Institute.
# 
# This software is provided by Ka-Ping Yee under the following
# license.  By using, copying, or modifying this software, you
# agree that you understand and will comply with the following
# terms and conditions: permission to use, copy, modify, and
# distribute this software and documentation for any purpose
# is granted, PROVIDED (a) that the full text of this notice
# appears on all copies of the software and documentation or
# portions or derivations thereof, and (b) that the preceding
# paragraph, containing the copyright, is displayed (together
# with other copyright notices if any) by all products that
# incorporate this software or portions or derivations thereof.
# If possible, we prefer that the URL be an active hyperlink.
# 
# THIS SOFTWARE IS PROVIDED "AS IS".  KA-PING YEE AND FORESIGHT
# INSTITUTE MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
# IMPLIED, INCLUDING, BUT NOT LIMITED TO, REPRESENTATIONS OR
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
# PURPOSE OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY
# THIRD-PARTY PATENTS, COPYRIGHTS, TRADEMARKS, OR OTHER RIGHTS.
# NEITHER KA-PING YEE NOR FORESIGHT INSTITUTE WILL BEAR ANY
# LIABILITY FOR ANY USE OF THIS SOFTWARE OR DOCUMENTATION.
# 
# The name and trademarks of Ka-Ping Yee and the Foresight
# Institute may not be used in publicity about this software
# without prior written permission (please contact us at
# <ping@foresight.org>).  Title to copyright in this software
# and documentation will at all times remain with the author.

package ftpput;
use M1config;
use M1term;
use FileHandle;
use Socket;
use strict;

# Store a file in a particular account on an FTP server.
# Return nothing on success, descriptive message on error.
sub store
    {
    my ($server,                       # string: name or address of server
        $login,                        # string: FTP username
        $password,                     # string: FTP password
        $path,                         # string: path to store at
        $text) = @_;                   # string: data content to store

    my $port = 21;

    my ($name, $alias, $type, $len, $addr, $lastresponse);
    my $proto = (getprotobyname('tcp'))[2];

    # look up addresses and build the sockaddr structures
    &M1term::msg("looking up $server");
    if ($server =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) # numeric IP address
        { $addr = pack('C4', $1, $2, $3, $4); }
    else
        { ($name, $alias, $type, $len, $addr) = gethostbyname($server); }

    if (!$addr) { return "Could not resolve hostname $server."; }

    my $client = pack('S n a4 x8', AF_INET, 0, $M1config::localip);
    my $servctl = pack('S n a4 x8', AF_INET, $port, $addr);

    # open a control connection
    socket(CONTROL, PF_INET, SOCK_STREAM, $proto) ||
        return "Could not create a client socket.";
    bind(CONTROL, $client) ||
        return "Could not bind the client socket.";

    &M1term::msg("connect " . join('.',unpack('C4', $addr)) . ":$port");
    connect(CONTROL, $servctl) ||
        return "Could not open a control connection.";
    CONTROL->autoflush(1);
    if (&response() != 220) { return "Server did not respond."; }

    # log in on the control connection
    &M1term::msg("attempting login $login $password");
    &command("USER $login");
    if (&response() != 331) { return "Login failed."; }
    &command("PASS $password");
    if (&response() != 230) { return "Login failed."; }
    &command("CWD");
    if (&response() != 250) { return "Path is invalid."; }
    &command("TYPE I");
    if (&response() != 200) { return "Could not set type to binary."; }

    # seek to the specified path, creating directories along the way
    $path =~ s#^/*##;
    while ($path =~ s#^([^/]+)/##)
        {
        my $dir = $1;
        &command("MKD $dir");
        if (&response() !~ /550|257/) { return "Path is invalid."; }
        &command("CWD $dir");
        if (&response() != 250) { return "Path is invalid."; }
        }

    # put the server in passive mode and send the data to it
    &command("PASV");
    if (&response =~ /^227.*\D(\d+),(\d+),(\d+),(\d+),(\d+),(\d+)/)
        {
        my $daddr = pack('C4', $1, $2, $3, $4);
        my $dport = ($5 * 256) + $6;
        my $servdat = pack('S n a4 x8', AF_INET, $dport, $daddr);

        socket(DATA, PF_INET, SOCK_STREAM, $proto) ||
            return "Could not create a client socket.";
        bind(DATA, $client) ||
            return "Could not bind the client socket.";

        &command("STOR $path");

        &M1term::msg("connect " . join('.',unpack('C4', $daddr)) . ":$dport");
        connect(DATA, $servdat) || return "Could not open a data connection.";
        DATA->autoflush(1);
        print DATA $text;
        close(DATA);

        if (&response() != 150) { return "Server did not accept data."; }
        my $resp=&response();
        if ($resp != 226) { return "Transfer did not complete. $resp"; }
        &M1term::msg("data successfully transferred to $path");
        }
    else
        { return "Server did not support passive mode."; }
    close(CONTROL);
    return undef;

    sub command                        # send a command to the FTP server
        {
        print CONTROL "$_[0]\r\n";
        &M1term::dbg("command: $_[0]");
        }

    sub response                       # read a response line from the socket
        {
        $lastresponse = '';

        while (1)
            {
            my ($rfd, $wfd, $efd) = ('', '', '');
            my $sockfd = fileno(CONTROL);
            vec($rfd, $sockfd, 1) = 1;
            vec($efd, $sockfd, 1) = 1;

            my ($found, $left) = select($rfd, $wfd, $efd,
                                        $M1config::clienttimeout);
            if (!vec($rfd, $sockfd, 1)) { return undef; }

            my $chunk;
            my $bytes = sysread(CONTROL, $chunk, 4096);
            $lastresponse .= $chunk;
            if ($lastresponse =~ s/\n//)
                {
                &M1term::dbg("response: $lastresponse");
                return $lastresponse;
                }
            }
        }
    }

1;
